#include <caml/mlvalues.h>
#include <caml/memory.h>
#include <caml/alloc.h>
#include <caml/fail.h>
#include <caml/callback.h>
#include "glfw.h"

static int value2winparam(value v)
{
    switch(Int_val(v))
    {
	case 0: return GLFW_OPENED;
	case 1: return GLFW_ACTIVE;
	case 2: return GLFW_ICONIFIED;
	case 3: return GLFW_ACCELERATED;
	case 4: return GLFW_RED_BITS;
	case 5: return GLFW_GREEN_BITS;
	case 6: return GLFW_BLUE_BITS;
	case 7: return GLFW_ALPHA_BITS;
	case 8: return GLFW_DEPTH_BITS;
	case 9: return GLFW_STENCIL_BITS;
	case 10: return GLFW_REFRESH_RATE;
	case 11: return GLFW_ACCUM_RED_BITS;
	case 12: return GLFW_ACCUM_GREEN_BITS;
	case 13: return GLFW_ACCUM_BLUE_BITS;
	case 14: return GLFW_ACCUM_ALPHA_BITS;
	case 15: return GLFW_AUX_BUFFERS;
	case 16: return GLFW_STEREO;
	default: failwith("value2winparam: argument out of range");
    }
    return 0;
}

CAMLprim void ml_glfwSetGLlib(value name)
{
    CAMLparam1(name);
    _glfwSetGLlib(String_val(name));
    CAMLreturn0;
}

CAMLprim value ml_glfwInit(value unit)
{
    CAMLparam0();
    int i = glfwInit();
    CAMLreturn(Val_bool(i));
}

CAMLprim void ml_glfwTerminate(value unit)
{
    CAMLparam0();
    glfwTerminate();
    CAMLreturn0;
}

CAMLprim value ml_glfwGetVersion(value unit)
{
    CAMLparam0();
    CAMLlocal1(r);
    int a,b,c;
    glfwGetVersion(&a,&b,&c);
    r=alloc_tuple(3);
    Store_field(r,0,Val_int(a));
    Store_field(r,1,Val_int(b));
    Store_field(r,2,Val_int(c));
    CAMLreturn(r);
}

CAMLprim value ml_glfwOpenWindow(value w,value h,value rb,value gb,value bb,value ab,value db,value sb,value m)
{
    CAMLparam5(w,h,rb,gb,bb);
    CAMLxparam4(ab,db,sb,m);
    int i=glfwOpenWindow(Int_val(w),Int_val(h),Int_val(rb),Int_val(gb),Int_val(bb),Int_val(ab),Int_val(db),Int_val(sb),Int_val(m)?GLFW_FULLSCREEN:GLFW_WINDOW);
    CAMLreturn(Val_bool(i));
}

CAMLprim value byte_ml_glfwOpenWindow(value* argv,int argc)
{
    return ml_glfwOpenWindow(argv[0],argv[1],argv[2],argv[3],argv[4],argv[5],argv[6],argv[7],argv[8]);
}   

CAMLprim void ml_glfwOpenWindowHint(value t,value h)
{
    CAMLparam2(t,h);
    glfwOpenWindowHint(value2winparam(t),Int_val(h));
    CAMLreturn0;
}

CAMLprim void ml_glfwCloseWindow(value unit)
{
    CAMLparam0();
    glfwCloseWindow();
    CAMLreturn0;
}

CAMLprim void ml_glfwSetWindowTitle(value s)
{
    CAMLparam1(s);
    glfwSetWindowTitle(String_val(s));
    CAMLreturn0;
}

CAMLprim value ml_glfwGetWindowSize(value unit)
{
    CAMLparam0();
    CAMLlocal1(r);
    int a,b;
    glfwGetWindowSize(&a,&b);
    r=alloc_tuple(2);
    Store_field(r,0,a);
    Store_field(r,1,b);
    CAMLreturn(r);
}

CAMLprim void ml_glfwSetWindowSize(value a,value b)
{
    CAMLparam2(a,b);
    glfwSetWindowSize(Int_val(a),Int_val(b));
    CAMLreturn0;
}

CAMLprim void ml_glfwSetWindowPos(value a,value b)
{
    CAMLparam2(a,b);
    glfwSetWindowPos(Int_val(a),Int_val(b));
    CAMLreturn0;
}

CAMLprim void ml_glfwIconifyWindow(value unit)
{
    CAMLparam0();
    glfwIconifyWindow();
    CAMLreturn0;
}

CAMLprim void ml_glfwRestoreWindow(value unit)
{
    CAMLparam0();
    glfwRestoreWindow();
    CAMLreturn0;
}

CAMLprim void ml_glfwSwapBuffers(value unit)
{
    CAMLparam0();
    glfwSwapBuffers();
    CAMLreturn0;
}

CAMLprim void ml_glfwSwapInterval(value i)
{
    CAMLparam1(i);
    glfwSwapInterval(Int_val(i));
    CAMLreturn0;
}

CAMLprim value ml_glfwGetWindowParam(value i)
{
    CAMLparam1(i);
    int a=glfwGetWindowParam(value2winparam(i));
    CAMLreturn(Val_int(a));
}

CAMLprim value ml_glfwGetDesktopMode(value unit)
{
    CAMLparam0();
    CAMLlocal1(r);
    GLFWvidmode d;
    glfwGetDesktopMode(&d);
    r=alloc(5,0);
    Store_field(r,0,Val_int(d.Width));
    Store_field(r,1,Val_int(d.Height));
    Store_field(r,2,Val_int(d.RedBits));
    Store_field(r,3,Val_int(d.GreenBits));
    Store_field(r,4,Val_int(d.BlueBits));
    CAMLreturn(r);
}

CAMLprim value ml_glfwGetVideoModes(value unit)
{
    CAMLparam0();
    CAMLlocal2(r,t);
    int i;
    GLFWvidmode* v=(GLFWvidmode*)malloc(sizeof(GLFWvidmode)*1024);
    i=glfwGetVideoModes(v,1024);
    r=alloc(i,0);
    while(i>0)
    {
	i-=1;
	t=alloc(5,0);
	Store_field(t,0,Val_int(v[i].Width));
	Store_field(t,1,Val_int(v[i].Height));
	Store_field(t,2,Val_int(v[i].RedBits));
	Store_field(t,3,Val_int(v[i].GreenBits));
	Store_field(t,4,Val_int(v[i].BlueBits));
	modify(&Field(r,i),t);
    }
    free(v);
    CAMLreturn(r);
}

CAMLprim void ml_glfwPollEvents(value unit)
{
    CAMLparam0();
    glfwPollEvents();
    CAMLreturn0;
}

CAMLprim value ml_glfwGetKey(value key)
{
    CAMLparam1(key);
    int i = glfwGetKey(Int_val(key));
    CAMLreturn(Val_bool(i));
}

CAMLprim value ml_glfwGetMouseButton(value key)
{
    CAMLparam1(key);
    int i = glfwGetMouseButton(Int_val(key));
    CAMLreturn(Val_bool(i));
}

CAMLprim value ml_glfwGetMousePos(value unit)
{
    CAMLparam0();
    CAMLlocal1(r);
    int a,b;
    glfwGetMousePos(&a,&b);
    r=alloc_tuple(2);
    Store_field(r,0,a);
    Store_field(r,1,b);
    CAMLreturn(r);
}

CAMLprim void ml_glfwSetMousePos(value a,value b)
{
    CAMLparam2(a,b);
    glfwSetMousePos(Int_val(a),Int_val(b));
    CAMLreturn0;
}

CAMLprim value ml_glfwGetMouseWheel(value unit)
{
    CAMLparam0();
    int i=glfwGetMouseWheel();
    CAMLreturn(Val_int(i));
}

CAMLprim void ml_glfwSetMouseWheel(value i)
{
    CAMLparam1(i);
    glfwSetMouseWheel(i);
    CAMLreturn0;
}

CAMLprim value ml_glfwGetJoystickParam(value a,value b)
{
    CAMLparam2(a,b);
    int i = glfwGetJoystickParam(Int_val(a),Int_val(b)+0x00050001);
    CAMLreturn(Val_int(i));
}

CAMLprim value ml_glfwGetJoystickPos(value j,value a)
{
#ifdef __linux__
    CAMLparam2(j,a);
    CAMLlocal1(r);
    float f[Int_val(a)];
    int k;
    int i = glfwGetJoystickPos(Int_val(j),f,Int_val(a));
    r=alloc(i * Double_wosize, Double_array_tag);
    for(k=0;k<i;k+=1) Store_double_field(r,k,f[k]);
    CAMLreturn(r);
#elif _WIN32
    CAMLparam2(j,a);
    CAMLlocal1(r);
    float *f;
    int k;
    int i;
    f=(float*)malloc(sizeof(float)*Int_val(a));
    i = glfwGetJoystickPos(Int_val(j),f,Int_val(a));
    r=alloc(i * Double_wosize, Double_array_tag);
    for(k=0;k<i;k+=1) Store_double_field(r,k,f[k]);
    free(f);
    CAMLreturn(r);
#else
#error "not linux or windows"
#endif
}

CAMLprim value ml_glfwGetJoystickButtons(value j,value n)
{
#ifdef __linux__
    CAMLparam2(j,n);
    CAMLlocal1(r);
    unsigned char b[Int_val(n)];
    int k;
    int i=glfwGetJoystickButtons(Int_val(j),b,Int_val(n));
    r=alloc(i,0);
    for(k=0;k<i;k+=1) Store_field(r,k,Val_bool(b[k]));
    CAMLreturn(r);
#elif _WIN32
    CAMLparam2(j,n);
    CAMLlocal1(r);
    unsigned char* b;
    int k,i;
    b=(unsigned char*)malloc(Int_val(n));
    i=glfwGetJoystickButtons(Int_val(j),b,Int_val(n));
    r=alloc(i,0);
    for(k=0;k<i;k+=1) Store_field(r,k,Val_bool(b[k]));
    free(b);
    CAMLreturn(r);
#else
#error "not linux or windows"
#endif
}

CAMLprim void ml_glfwEnable(value i)
{
    CAMLparam1(i);
    glfwEnable(Int_val(i)+0x00030001);
    CAMLreturn0;
}

CAMLprim void ml_glfwDisable(value i)
{
    CAMLparam1(i);
    glfwDisable(Int_val(i)+0x00030001);
    CAMLreturn0;
}

static void __protoswsc(int a,int b)
{
    callback2(*caml_named_value("ml_setwindowsizecallback"),Val_int(a),Val_int(b));
}

CAMLprim void ml_glfwProtoswsc(value unit)
{
    CAMLparam0();
    glfwSetWindowSizeCallback(__protoswsc);
    CAMLreturn0;
}

static void __protoskc(int a,int b)
{
    callback2(*caml_named_value("ml_setkeycallback"),Val_int(a),Val_bool(b));
}

CAMLprim void ml_glfwProtoskc(value unit)
{
    CAMLparam0();
    glfwSetKeyCallback(__protoskc);
    CAMLreturn0;
}

static void __protombc(int a,int b)
{
    callback2(*caml_named_value("ml_setmousebuttoncallback"),Val_int(a),Val_bool(b));
}

CAMLprim void ml_glfwProtombc(value unit)
{
    CAMLparam0();
    glfwSetMouseButtonCallback(__protombc);
    CAMLreturn0;
}

static void __protompc(int a,int b)
{
    callback2(*caml_named_value("ml_setmouseposcallback"),Val_int(a),Val_int(b));
}

CAMLprim void ml_glfwProtompc(value unit)
{
    CAMLparam0();
    glfwSetMousePosCallback(__protompc);
    CAMLreturn0;
}

static void __protomwc(int a)
{
    callback(*caml_named_value("ml_setmousewheelcallback"),Val_int(a));
}

CAMLprim void ml_glfwProtomwc(value unit)
{
    CAMLparam0();
    glfwSetMouseWheelCallback(__protomwc);
    CAMLreturn0;
}

