//========================================================================
// GLFW - An OpenGL framework
// File:     init.c
// Platform: Windows
// Version:  2.3
// Date:     2002.11.23
// Author:   Marcus Geelnard (marcus.geelnard@home.se)
// WWW:      http://hem.passagen.se/opengl/glfw/
//------------------------------------------------------------------------
// Copyright (c) 2002 Marcus Geelnard
//
// GLFW is provided "as-is", without any express or implied warranty,
// without even the implied warranty of fitness for a particular purpose.
// In no event will the author be held liable for any damage araising from
// the use of GLFW.
//
// Permission is granted to anyone to use GLFW for any purpose, including
// commercial applications, and to alter it and redistribute it freely,
// subject to the following restrictions:
//
// 1. You must not claim that you wrote GLFW.
// 2. If you modify any parts of GLFW, it must be clearly stated which
//    modifications have been made, and that those modifications are not
//    the work of the author of GLFW.
// 3. If you use GLFW in a product, an acknowledgment in the product
//    documentation would be appreciated, but is not required.
// 4. This notice may not be removed or altered from any source
//    distribution of GLFW.
//
// Marcus Geelnard
// marcus.geelnard@home.se
//========================================================================

#define _init_c_
#include "internal.h"

// With the Borland C++ compiler, we want to disable FPU exceptions
#ifdef __BORLANDC__
#include <float.h>
#endif // __BORLANDC__



//************************************************************************
//****                  GLFW internal functions                       ****
//************************************************************************

//========================================================================
// _glfwInitLibraries() - Load necessary libraries (DLLs)
//========================================================================

static int _glfwInitLibraries( void )
{
    // gdi32.dll (OpenGL pixel format functions & SwapBuffers)
#ifndef GLFW_NO_DLOAD_GDI32
    _glfwLibs.gdi32 = LoadLibrary( "gdi32.dll" );
    if( _glfwLibs.gdi32 != NULL )
    {
        _glfwLibs.ChoosePixelFormat   = (CHOOSEPIXELFORMAT_T)
            GetProcAddress( _glfwLibs.gdi32, "ChoosePixelFormat" );
        _glfwLibs.DescribePixelFormat = (DESCRIBEPIXELFORMAT_T)
            GetProcAddress( _glfwLibs.gdi32, "DescribePixelFormat" );
        _glfwLibs.GetPixelFormat      = (GETPIXELFORMAT_T)
            GetProcAddress( _glfwLibs.gdi32, "GetPixelFormat" );
        _glfwLibs.SetPixelFormat      = (SETPIXELFORMAT_T)
            GetProcAddress( _glfwLibs.gdi32, "SetPixelFormat" );
        _glfwLibs.SwapBuffers         = (SWAPBUFFERS_T)
            GetProcAddress( _glfwLibs.gdi32, "SwapBuffers" );
        if( _glfwLibs.ChoosePixelFormat   == NULL ||
            _glfwLibs.DescribePixelFormat == NULL ||
            _glfwLibs.GetPixelFormat      == NULL ||
            _glfwLibs.SetPixelFormat      == NULL ||
            _glfwLibs.SwapBuffers         == NULL )
        {
            return GL_FALSE;
        }
    }
    else
    {
        return GL_FALSE;
    }
#endif // GLFW_NO_DLOAD_GDI32

    // winmm.dll (for joystick support)
#ifndef GLFW_NO_DLOAD_WINMM
    _glfwLibs.winmm = LoadLibrary( "winmm.dll" );
    if( _glfwLibs.winmm != NULL )
    {
        _glfwLibs.joyGetDevCapsA = (JOYGETDEVCAPSA_T)
            GetProcAddress( _glfwLibs.winmm, "joyGetDevCapsA" );
        _glfwLibs.joyGetPos      = (JOYGETPOS_T)
            GetProcAddress( _glfwLibs.winmm, "joyGetPos" );
        _glfwLibs.joyGetPosEx    = (JOYGETPOSEX_T)
            GetProcAddress( _glfwLibs.winmm, "joyGetPosEx" );
        if( _glfwLibs.joyGetDevCaps == NULL ||
            _glfwLibs.joyGetPos     == NULL ||
            _glfwLibs.joyGetPosEx   == NULL )
        {
            return GL_FALSE;
        }
    }
    else
    {
        return GL_FALSE;
    }
#endif // GLFW_NO_DLOAD_WINMM

    return GL_TRUE;
}


//========================================================================
// _glfwFreeLibraries() - Unload used libraries (DLLs)
//========================================================================

static void _glfwFreeLibraries( void )
{
    // gdi32.dll
#ifndef GLFW_NO_DLOAD_GDI32
    if( _glfwLibs.gdi32 != NULL )
    {
        FreeLibrary( _glfwLibs.gdi32 );
        _glfwLibs.gdi32 = NULL;
    }
#endif // GLFW_NO_DLOAD_GDI32

    // winmm.dll
#ifndef GLFW_NO_DLOAD_WINMM
    if( _glfwLibs.winmm != NULL )
    {
        FreeLibrary( _glfwLibs.winmm );
        _glfwLibs.winmm = NULL;
    }
#endif // GLFW_NO_DLOAD_WINMM
}

//========================================================================
// _glfwTerminate() - Called from glfwTerminate()
//========================================================================

void _glfwTerminate( void )
{
    // Is GLFW initialised?
    if( !_glfwInitialised )
    {
        return;
    }

    // Close OpenGL window
    glfwCloseWindow();

    // Enable system keys again (if they were disabled)
    glfwEnable( GLFW_SYSTEM_KEYS );

    // Unload libraries (DLLs)
    _glfwFreeLibraries();

    // Restore FOREGROUNDLOCKTIMEOUT system setting
    SystemParametersInfo( SPI_SETFOREGROUNDLOCKTIMEOUT, 0,
                          (LPVOID)_glfwSys.ForegroundLockTimeout,
                          SPIF_SENDCHANGE );

    // GLFW is no longer initialised
    _glfwInitialised = GL_FALSE;
}


//************************************************************************
//****                    GLFW user functions                         ****
//************************************************************************

//========================================================================
// glfwInit() - Initialize various GLFW state
//========================================================================

GLFWAPI int GLFWAPIENTRY glfwInit( void )
{
    OSVERSIONINFO osi;

    // Is GLFW already running?
    if( _glfwInitialised )
    {
        return GL_TRUE;
    }

    // To make SetForegroundWindow() work as we want, we need to fiddle
    // with the FOREGROUNDLOCKTIMEOUT system setting (we do this as early
    // as possible in the hope of still being the foreground process)
    SystemParametersInfo( SPI_GETFOREGROUNDLOCKTIMEOUT, 0,
                          &_glfwSys.ForegroundLockTimeout, 0 );
    SystemParametersInfo( SPI_SETFOREGROUNDLOCKTIMEOUT, 0, (LPVOID)0,
                          SPIF_SENDCHANGE );

    // Load libraries (DLLs)
    if( !_glfwInitLibraries() )
    {
        return GL_FALSE;
    }

    // With the Borland C++ compiler, we want to disable FPU exceptions
    // (this is recommended for OpenGL applications under Windows)
#ifdef __BORLANDC__
    _control87( MCW_EM, MCW_EM );
#endif

    // Check which OS version we are running
    osi.dwOSVersionInfoSize = sizeof( OSVERSIONINFO );
    GetVersionEx( &osi );
    _glfwSys.WinVer = _GLFW_WIN_UNKNOWN;
    if( osi.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS )
    {
        if( osi.dwMajorVersion == 4 && osi.dwMinorVersion < 10 )
        {
            _glfwSys.WinVer = _GLFW_WIN_95;
        }
        else if( osi.dwMajorVersion == 4 && osi.dwMinorVersion < 90 )
        {
            _glfwSys.WinVer = _GLFW_WIN_98;
        }
        else if( osi.dwMajorVersion == 4 && osi.dwMinorVersion == 90 )
        {
            _glfwSys.WinVer = _GLFW_WIN_ME;
        }
        else if( osi.dwMajorVersion >= 4 )
        {
            _glfwSys.WinVer = _GLFW_WIN_UNKNOWN_9x;
        }
    }
    else if( osi.dwPlatformId == VER_PLATFORM_WIN32_NT )
    {
        if( osi.dwMajorVersion == 4 && osi.dwMinorVersion == 0 )
        {
            _glfwSys.WinVer = _GLFW_WIN_NT4;
        }
        else if( osi.dwMajorVersion == 5 && osi.dwMinorVersion == 0 )
        {
            _glfwSys.WinVer = _GLFW_WIN_2K;
        }
        else if( osi.dwMajorVersion == 5 && osi.dwMinorVersion == 1 )
        {
            _glfwSys.WinVer = _GLFW_WIN_XP;
        }
        else if( osi.dwMajorVersion == 5 && osi.dwMinorVersion == 2 )
        {
            _glfwSys.WinVer = _GLFW_WIN_NET_SERVER;
        }
        else if( osi.dwMajorVersion >= 5 )
        {
            _glfwSys.WinVer = _GLFW_WIN_UNKNOWN_NT;
        }
    }

    // Window is not yet opened
    _glfwWin.Opened = GL_FALSE;

    // System keys are not disabled
    _glfwWin.KeyboardHook = NULL;
    _glfwWin.SysKeysDisabled = GL_FALSE;

    // Inactivate OpenGL extension WGL_EXT_swap_control (activated by
    // glfwOpenWindow)
    _wglSwapIntervalEXT = (WGLSWAPINTERVALEXT_T) 0;

    // Clear window hints
    _glfwWinHints.RefreshRate    = 0;
    _glfwWinHints.AccumRedBits   = 0;
    _glfwWinHints.AccumGreenBits = 0;
    _glfwWinHints.AccumBlueBits  = 0;
    _glfwWinHints.AccumAlphaBits = 0;
    _glfwWinHints.AuxBuffers     = 0;
    _glfwWinHints.Stereo         = 0;

    // Install atexit() routine
    atexit( _glfwTerminate );

    // Form now on, GLFW state is valid
    _glfwInitialised = GL_TRUE;

    return GL_TRUE;
}


//========================================================================
// glfwTerminate() - Close window and kill all threads.
//========================================================================

GLFWAPI void GLFWAPIENTRY glfwTerminate( void )
{
    _glfwTerminate();
}


//========================================================================
// glfwGetVersion() - Get GLFW version
//========================================================================

GLFWAPI void GLFWAPIENTRY glfwGetVersion( int *major, int *minor,
    int *rev )
{
    if( major != NULL ) *major = GLFW_VERSION_MAJOR;
    if( minor != NULL ) *minor = GLFW_VERSION_MINOR;
    if( rev   != NULL ) *rev   = GLFW_VERSION_REVISION;
}
