#include <caml/mlvalues.h>
#include <caml/alloc.h>
#include <caml/memory.h>
#include <caml/bigarray.h>
#include <caml/fail.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

value ba_float_set(value size,float data)
{
	float* a;
	int i,s;
	s=Int_val(size);
	a=(float*)malloc(s*sizeof(float));
	assert(a);
	for(i=0;i<s;i+=1) a[i]=data; /* Todo: optimize */
	return alloc_bigarray_dims(BIGARRAY_FLOAT32|BIGARRAY_C_LAYOUT|BIGARRAY_MANAGED,1,a,s);	
}

value ba_double_set(value size,double data)
{
	double* a;
	int i,s;
	s=Int_val(size);
	a=(double*)malloc(s*sizeof(double));
	assert(a);
	for(i=0;i<s;i+=1) a[i]=data; /* Todo: optimize */
	return alloc_bigarray_dims(BIGARRAY_FLOAT64|BIGARRAY_C_LAYOUT|BIGARRAY_MANAGED,1,a,s);	
}

value c_proml_types_int_set(value size,value data) /* TODO */
{
	int* a;
	int i,s,v;
	s=Int_val(size);
	a=(int*)malloc(s*sizeof(int));
	assert(a);
	v=Int_val(data);
	for(i=0;i<s;i+=1) a[i]=v; /* Todo: optimize */
	return alloc_bigarray_dims(BIGARRAY_CAML_INT|BIGARRAY_C_LAYOUT|BIGARRAY_MANAGED,1,a,s);	
}

value c_proml_types_byte_set(value size,value data) /* TODO */
{
	unsigned char* a;
	int i,s;
	unsigned char v;
	s=Int_val(size);
	a=(unsigned char*)malloc(s*sizeof(unsigned char));
	assert(a);
	v=Int_val(data);
	for(i=0;i<s;i+=1) a[i]=v; /* Todo: optimize */
	return alloc_bigarray_dims(BIGARRAY_UINT8|BIGARRAY_C_LAYOUT|BIGARRAY_MANAGED,1,a,s);	
}

value c_proml_types_byte_rgb(value size,value vr,value vg,value vb)
{
	unsigned char* a;
	unsigned char r,g,b;
	int i,s;
	s=Int_val(size)*3;
	a=(unsigned char*)malloc(s*sizeof(unsigned char));
	assert(a);
	r=Int_val(vr);
	g=Int_val(vg);
	b=Int_val(vb);
	for(i=0;i<s;i+=3) /* TODO: optimize */
	{
		a[i+2]=b;
		a[i+1]=g;
		a[i]=r;
	}
	return alloc_bigarray_dims(BIGARRAY_UINT8|BIGARRAY_C_LAYOUT|BIGARRAY_MANAGED,1,a,s);	
}

value c_proml_types_byte_rgba(value size,value vr,value vg,value vb,value va)
{
	unsigned char* a;
	unsigned char r,g,b,ca;
	int i,s;
	s=Int_val(size)*4;
	a=(unsigned char*)malloc(s*sizeof(unsigned char));
	assert(a);
	r=Int_val(vr);
	g=Int_val(vg);
	b=Int_val(vb);
	ca=Int_val(va);
	for(i=0;i<s;i+=4) /* TODO: optimize */
	{
		a[i+3]=ca;
		a[i+2]=b;
		a[i+1]=g;
		a[i]=r;
	}
	return alloc_bigarray_dims(BIGARRAY_UINT8|BIGARRAY_C_LAYOUT|BIGARRAY_MANAGED,1,a,s);	
}

	
value c_proml_types_float_set(value size,value data) /* TODO */
{
	return ba_float_set(size,Double_val(data));
}

value c_proml_types_double_set(value size,value data) 
{
	return ba_double_set(size,Double_val(data));
}

value c_proml_types_int_zero(value size) 
{
	return c_proml_types_int_set(size,Val_int(0));
}

value c_proml_types_float_zero(value size) 
{
	return ba_float_set(size,0.0);
}

value c_proml_types_double_zero(value size) 
{
	return ba_double_set(size,0.0);
}

value c_proml_types_byte_zero(value size) 
{
	return c_proml_types_byte_set(size,Val_int(0));;
}

value c_proml_types_float_init(value arr)
{
	CAMLparam1(arr);
	CAMLlocal1(ba);
	float* a;
	int size=Wosize_val(arr)>>1; /* possible bug */
	int i;
	a=(float*)malloc(size*sizeof(float));
	if(a==0) failwith("c_proml_types_float_init: malloc failed");
	for(i=0;i<size;i+=1) a[i]=Double_field(arr,i);
	ba=alloc_bigarray_dims(BIGARRAY_FLOAT32|BIGARRAY_C_LAYOUT|BIGARRAY_MANAGED,1,a,size);
	CAMLreturn(ba);
}

value c_proml_types_double_init(value arr)
{
	CAMLparam1(arr);
	CAMLlocal1(ba);
	double* a;
	int size=Wosize_val(arr)>>1; /* possible bug */
	int i;
	a=(double*)malloc(size*sizeof(double));
	if(a==0) failwith("c_proml_types_double_init: malloc failed");
	for(i=0;i<size;i+=1) a[i]=Double_field(arr,i);
	ba=alloc_bigarray_dims(BIGARRAY_FLOAT64|BIGARRAY_C_LAYOUT|BIGARRAY_MANAGED,1,a,size);
	CAMLreturn(ba);
}

value c_proml_types_int_init(value arr)
{
	CAMLparam1(arr);
	CAMLlocal1(ba);
	int* a;
	int size=Wosize_val(arr); /* possible bug */
	int i;
	a=(int*)malloc(size*sizeof(int));
	if(a==0) failwith("c_proml_types_int_init: malloc failed");
	for(i=0;i<size;i+=1) a[i]=Int_val(Field(arr,i));
	ba=alloc_bigarray_dims(BIGARRAY_CAML_INT|BIGARRAY_C_LAYOUT|BIGARRAY_MANAGED,1,a,size);
	CAMLreturn(ba);
}

value c_proml_types_byte_init(value arr)
{
	CAMLparam1(arr);
	CAMLlocal1(ba);
	unsigned char* a;
	int size=Wosize_val(arr); /* possible bug */
	int i;
	a=(unsigned char*)malloc(size*sizeof(unsigned char));
	if(a==0) failwith("c_proml_types_byte_init: malloc failed");
	for(i=0;i<size;i+=1) a[i]=Int_val(Field(arr,i));
	ba=alloc_bigarray_dims(BIGARRAY_UINT8|BIGARRAY_C_LAYOUT|BIGARRAY_MANAGED,1,a,size);
	CAMLreturn(ba);
}

static int typesize(int t)
{
    switch(t)
    {
	case BIGARRAY_FLOAT32: return 4;
	case BIGARRAY_FLOAT64: return 8;
	case BIGARRAY_SINT8: return 1;
	case BIGARRAY_UINT8: return 1;
	case BIGARRAY_SINT16: return 2;
	case BIGARRAY_UINT16: return 2;
	case BIGARRAY_INT32: return 4;
	case BIGARRAY_INT64: return 8;
	case BIGARRAY_CAML_INT: 
	case BIGARRAY_NATIVE_INT: 
#ifdef ARCH_SIXTYFOUR
		return 8;
#else
		return 4;
#endif	
	case BIGARRAY_COMPLEX32: return 8;
	case BIGARRAY_COMPLEX64: return 16;
	default: failwith("typesize: unknown Bigarray type");
    }
    return 0;
}

value ml__baofs(value ba,value vofs,value vsize,int type)
{
    char* v=(char*)Data_bigarray_val(ba);
    value r;
    int size=Int_val(vsize);
    int ofs=Int_val(vofs);
    int rsize=Bigarray_val(ba)->dim[0] * typesize(Bigarray_val(ba)->flags & BIGARRAY_KIND_MASK);
    if(rsize<ofs+size*typesize(type)) 
	failwith("ml__baofs: out of range");
    v+=ofs;
    r=alloc_bigarray_dims(type|BIGARRAY_C_LAYOUT,1,v,size);
    return r;
}

value ml__baofs_float(value ba,value vofs,value vsize)
{
    CAMLparam3(ba,vofs,vsize);
    CAMLlocal1(r);
    r=ml__baofs(ba,vofs,vsize,BIGARRAY_FLOAT32);
    CAMLreturn(r);
}

value ml__baofs_double(value ba,value vofs,value vsize)
{
    CAMLparam3(ba,vofs,vsize);
    CAMLlocal1(r);
    r=ml__baofs(ba,vofs,vsize,BIGARRAY_FLOAT64);
    CAMLreturn(r);
}

value ml__baofs_uint8(value ba,value vofs,value vsize)
{
    CAMLparam3(ba,vofs,vsize);
    CAMLlocal1(r);
    r=ml__baofs(ba,vofs,vsize,BIGARRAY_UINT8);
    CAMLreturn(r);
}

value ml__baofs_sint8(value ba,value vofs,value vsize)
{
    CAMLparam3(ba,vofs,vsize);
    CAMLlocal1(r);
    r=ml__baofs(ba,vofs,vsize,BIGARRAY_SINT8);
    CAMLreturn(r);
}

value ml__baofs_uint16(value ba,value vofs,value vsize)
{
    CAMLparam3(ba,vofs,vsize);
    CAMLlocal1(r);
    r=ml__baofs(ba,vofs,vsize,BIGARRAY_UINT16);
    CAMLreturn(r);
}

value ml__baofs_sint16(value ba,value vofs,value vsize)
{
    CAMLparam3(ba,vofs,vsize);
    CAMLlocal1(r);
    r=ml__baofs(ba,vofs,vsize,BIGARRAY_SINT16);
    CAMLreturn(r);
}

value ml__baofs_int(value ba,value vofs,value vsize)
{
    CAMLparam3(ba,vofs,vsize);
    CAMLlocal1(r);
    r=ml__baofs(ba,vofs,vsize,BIGARRAY_CAML_INT);
    CAMLreturn(r);
}

