#include <SDL/SDL.h>
#include <caml/mlvalues.h>
#include <caml/alloc.h>
#include <caml/memory.h>

#include "shared.h"

#define ML_surf_to_C(surf) 	( (SDL_Surface *)Field(surf, 0) )
#define ML_pixfmt_to_C(pixfmt)	( (SDL_PixelFormat *)Field(pixfmt, 0) )

static value SDL_pixfmt_to_ML_pixfmt(SDL_PixelFormat *pixfmt)
{
    CAMLparam0();
    CAMLlocal2(ML_pixfmt, ML_add);
    
    ML_pixfmt=alloc(9,0);
    Store_field(ML_pixfmt, 0, (value)(pixfmt));
    Store_field(ML_pixfmt, 1, (value)(pixfmt->palette));
    Store_field(ML_pixfmt, 2, Val_int(pixfmt->BitsPerPixel));
    Store_field(ML_pixfmt, 3, Val_int(pixfmt->BytesPerPixel));
	ML_add=alloc(4,0);
	Store_field(ML_add, 0, Val_int(pixfmt->Rloss));
	Store_field(ML_add, 1, Val_int(pixfmt->Gloss));
	Store_field(ML_add, 2, Val_int(pixfmt->Bloss));
	Store_field(ML_add, 3, Val_int(pixfmt->Aloss));
    Store_field(ML_pixfmt, 4, ML_add);
	ML_add=alloc(4,0);
	Store_field(ML_add, 0, Val_int(pixfmt->Rshift));
	Store_field(ML_add, 1, Val_int(pixfmt->Gshift));
	Store_field(ML_add, 2, Val_int(pixfmt->Bshift));
	Store_field(ML_add, 3, Val_int(pixfmt->Ashift));
    Store_field(ML_pixfmt, 5, ML_add);
	ML_add=alloc(4,0);
	Store_field(ML_add, 0, copy_int32(pixfmt->Rmask));
	Store_field(ML_add, 1, copy_int32(pixfmt->Gmask));
	Store_field(ML_add, 2, copy_int32(pixfmt->Bmask));
	Store_field(ML_add, 3, copy_int32(pixfmt->Amask));
    Store_field(ML_pixfmt, 6, ML_add);
    Store_field(ML_pixfmt, 7, copy_int32(pixfmt->colorkey));
    Store_field(ML_pixfmt, 8, copy_int32(pixfmt->alpha));
    CAMLreturn(ML_pixfmt);
}    

value SDL_surf_to_ML_surf(SDL_Surface *surf)
{
    CAMLparam0();
    CAMLlocal2(ML_surf,ML_add);
    
    ML_surf = alloc(15, 0);
    Store_field(ML_surf, 0, (value)(surf));
    Store_field(ML_surf, 1, copy_int32(surf->flags));
    Store_field(ML_surf, 2, SDL_pixfmt_to_ML_pixfmt(surf->format));
    Store_field(ML_surf, 3, Val_int(surf->w));
    Store_field(ML_surf, 4, Val_int(surf->h));
    Store_field(ML_surf, 5, Val_int(surf->pitch));
    Store_field(ML_surf, 6, (value)(surf->pixels));
    Store_field(ML_surf, 7, Val_int(surf->offset));
    Store_field(ML_surf, 8, (value)(surf->hwdata));
	ML_add=alloc(4, 0);
	Store_field(ML_add, 0, Val_int(surf->clip_rect.x));
	Store_field(ML_add, 1, Val_int(surf->clip_rect.y));
	Store_field(ML_add, 2, Val_int(surf->clip_rect.w));
	Store_field(ML_add, 3, Val_int(surf->clip_rect.h));
    Store_field(ML_surf, 9, ML_add);
    Store_field(ML_surf,10, copy_int32(surf->unused1));
    Store_field(ML_surf,11, copy_int32(surf->locked));
    Store_field(ML_surf,12, (value)(surf->map));
    Store_field(ML_surf,13, Val_int(surf->format_version));
    Store_field(ML_surf,14, Val_int(surf->refcount));
    
    CAMLreturn (ML_surf);
}


value MLSDL_mustlock(value surface)
{
    CAMLparam1(surface);
    CAMLreturn ( Val_bool( SDL_MUSTLOCK(ML_surf_to_C(surface)) ) );
}

void MLSDL_video_quit(value u)
{
    CAMLparam1(u);
    SDL_VideoQuit();
    CAMLreturn0;
}

value MLSDL_get_video_info(value u)
{
    CAMLparam1(u);
    const SDL_VideoInfo *info;
    CAMLlocal1(ML_info);
    
    info=SDL_GetVideoInfo();
    ML_info=alloc(11,0);
    Store_field(ML_info, 0, Val_bool(info->hw_available));
    Store_field(ML_info, 1, Val_bool(info->wm_available));
    Store_field(ML_info, 2, Val_bool(info->blit_hw));
    Store_field(ML_info, 3, Val_bool(info->blit_hw_CC));
    Store_field(ML_info, 4, Val_bool(info->blit_hw_A));
    Store_field(ML_info, 5, Val_bool(info->blit_sw));
    Store_field(ML_info, 6, Val_bool(info->blit_sw_CC));
    Store_field(ML_info, 7, Val_bool(info->blit_sw_A));
    Store_field(ML_info, 8, Val_bool(info->blit_fill));
    Store_field(ML_info, 9, Val_int(info->video_mem));
    Store_field(ML_info, 10, SDL_pixfmt_to_ML_pixfmt(info->vfmt));
    CAMLreturn(ML_info);
}

unsigned int flag_to_surftype [] =
{
    0x00000000,
    0x00000001,
    0x00000004,
    0x10000000,
    0x20000000,
    0x40000000,
    0x80000000,
    0x00000002,
    0x0000000A,
    0x00000010,
    0x00000020,
    0x00000100,
    0x00001000,
    0x00002000,
    0x00004000,
    0x00010000,
    0x01000000
};
#define FLAG_TO_SURFTYPE_SIZE 	17


value MLSDL_set_video_mode(value width, value height, value bpp, value flags)
{
    CAMLparam4(width, height, bpp, flags);
    SDL_Surface *surf;
    
    surf = SDL_SetVideoMode ( Int_val(width), Int_val(height), Int_val(bpp), 
	ML_flags_to_mask(flags, flag_to_surftype) );
    
    if (surf==NULL) fail();
    
    CAMLreturn (SDL_surf_to_ML_surf(surf));
}


void MLSDL_update_rect(value screen, value rect)
{
    CAMLparam2(screen, rect);
    SDL_UpdateRect( 
	ML_surf_to_C(screen), 
	Int_val(Field(rect,0)), 
	Int_val(Field(rect,1)), 
	Int_val(Field(rect,2)), 
	Int_val(Field(rect,3))
    );
    CAMLreturn0;
}

void MLSDL_flip(value screen)
{
    CAMLparam1(screen);
    if (SDL_Flip( ML_surf_to_C(screen) ) == -1)
	fail();
    CAMLreturn0;
}

void MLSDL_set_gamma(value rgb)
{
    CAMLparam1(rgb);
    if(SDL_SetGamma(
	Double_val(Field(rgb,0)), 
	Double_val(Field(rgb,1)), 
	Double_val(Field(rgb,2))
    )) fail();
  CAMLreturn0;
}

value MLSDL_map_RGB(value format, value rgb)
{
    CAMLparam2(format, rgb);
    CAMLreturn (copy_int32(
	SDL_MapRGB( 
	    ML_pixfmt_to_C(format), 
	    Int_val(Field(rgb,0)),
	    Int_val(Field(rgb,1)),
	    Int_val(Field(rgb,2))
	)
    ));
}

value MLSDL_map_RGBA(value format, value rgba)
{
    CAMLparam2(format, rgba);
    CAMLreturn (copy_int32(
	SDL_MapRGBA( 
	    ML_pixfmt_to_C(format),
	    Int_val(Field(rgba,0)),
	    Int_val(Field(rgba,1)),
	    Int_val(Field(rgba,2)),
	    Int_val(Field(rgba,3))
	)
    ));
}
    
value MLSDL_get_RGB(value pixel, value format)
{
    CAMLparam2(pixel, format);
    char r,g,b;
    CAMLlocal1(to_return);
    
    to_return=alloc_tuple(3);
    SDL_GetRGB( Int32_val(pixel), ML_pixfmt_to_C(format), &r, &g, &b);
    Store_field(to_return, 0, r);
    Store_field(to_return, 1, g);
    Store_field(to_return, 2, b);
    
    CAMLreturn (to_return);
}
    
value MLSDL_get_RGBA(value pixel, value format)
{
    CAMLparam2(pixel, format);
    char r,g,b,a;
    CAMLlocal1(to_return);
    to_return=alloc_tuple(4);
    
    SDL_GetRGBA( Int32_val(pixel), ML_pixfmt_to_C(format), &r, &g, &b, &a);
    Store_field(to_return, 0, r);
    Store_field(to_return, 1, g);
    Store_field(to_return, 2, b);
    Store_field(to_return, 3, a);

    CAMLreturn (to_return);
}

value MLSDL_create_RGB_surface(value flags, value whd, value RGBAmasks)
{
    CAMLparam3(flags, whd, RGBAmasks);
    SDL_Surface *surf;
    
    surf=SDL_CreateRGBSurface( ML_flags_to_mask(flags, flag_to_surftype),
	Int_val(Field(whd,0)), Int_val(Field(whd,1)), Int_val(Field(whd,1)),
	Int32_val(Field(RGBAmasks,0)), Int32_val(Field(RGBAmasks,1)), 
	Int32_val(Field(RGBAmasks,2)), Int32_val(Field(RGBAmasks,3))
    );
    if (surf==NULL) fail();
    CAMLreturn(SDL_surf_to_ML_surf(surf));
}

value MLSDL_create_RGB_surface_from(value pixels, value whd,
				    value pitch ,value RGBAmasks)
{
    CAMLparam4(pixels, whd, pitch, RGBAmasks);
    SDL_Surface *surf;
    
    surf=SDL_CreateRGBSurfaceFrom( (void *)pixels,
	Int_val(Field(whd,0)), Int_val(Field(whd,1)), Int_val(Field(whd,1)),
	Int_val(pitch),
	Int32_val(Field(RGBAmasks,0)), Int32_val(Field(RGBAmasks,1)), 
	Int32_val(Field(RGBAmasks,2)), Int32_val(Field(RGBAmasks,3))
    );
    if (surf==NULL) fail();
    CAMLreturn(SDL_surf_to_ML_surf(surf));
}

void MLSDL_free_surface(value surface)
{
    CAMLparam1(surface);
    SDL_FreeSurface( ML_surf_to_C(surface));
    CAMLreturn0;
}

void MLSDL_lock_surface(value surf)
{
    CAMLparam1(surf);
    if(SDL_LockSurface( ML_surf_to_C(surf) ) == -1)
	fail();
    CAMLreturn0;
}

void MLSDL_unlock_surface(value surf)
{
    CAMLparam1(surf);
    SDL_UnlockSurface( ML_surf_to_C(surf) );
    CAMLreturn0;
}


value MLSDL_load_BMP(value filename)
{
    CAMLparam1(filename);
    SDL_Surface *surf;
    
    surf = SDL_LoadBMP(String_val(filename));
    if(surf == NULL) fail();
    CAMLreturn (SDL_surf_to_ML_surf(surf));
}

void MLSDL_set_color_key(value surf, value flags, value key)
{
    CAMLparam3(surf, flags, key);
    if(
	SDL_SetColorKey(
	    ML_surf_to_C(surf),
	    ML_flags_to_mask(flags, flag_to_surftype),
	    Int32_val(key)
	)
    == -1) fail();
    CAMLreturn0;
}

void MLSDL_set_color_key_00(value surf, value flags)
{
    CAMLparam2(surf, flags);
    if(
	SDL_SetColorKey(
	    ML_surf_to_C(surf),
	    ML_flags_to_mask(flags, flag_to_surftype),
	    *(Uint8 *)(ML_surf_to_C(surf))->pixels
	)
    == -1) fail();
    CAMLreturn0;
}

void MLSDL_set_alpha(value surf, value flags, value alpha)
{
    CAMLparam3(surf, flags, alpha);
    if(
	SDL_SetAlpha(
	    ML_surf_to_C(surf),
	    ML_flags_to_mask(flags, flag_to_surftype),
	    Int_val(alpha)
	)
    == -1) fail();
    CAMLreturn0;
}

void MLSDL_set_clip_rect(value surf, value rect)
{
    CAMLparam2(surf, rect);
    CAMLlocal5(oldrect,x,y,w,h);
    SDL_Surface *csurf;
    x=Field(rect,0);
    y=Field(rect,1);
    w=Field(rect,2);
    h=Field(rect,3);
    
    oldrect=Field(surf, 9);
    Field(oldrect, 0)=x;
    Field(oldrect, 1)=y;
    Field(oldrect, 2)=w;
    Field(oldrect, 3)=h;
    csurf=ML_surf_to_C(surf);
    csurf->clip_rect.x=Int_val(x);
    csurf->clip_rect.y=Int_val(y);
    csurf->clip_rect.w=Int_val(w);
    csurf->clip_rect.h=Int_val(h);
    SDL_SetClipRect(csurf, &(csurf->clip_rect));
    CAMLreturn0;
}

value MLSDL_convert_surface(value src, value fmt, value flags)
{
    CAMLparam3(src, fmt, flags);
    SDL_Surface *surf;
    
    surf=SDL_ConvertSurface( ML_surf_to_C(src), ML_pixfmt_to_C(fmt), 
		    ML_flags_to_mask(flags, flag_to_surftype) );
    if(surf==NULL) fail();
    CAMLreturn(SDL_surf_to_ML_surf(surf));
}

void MLSDL_blit_surface(value src, value srcrect, value dst, value dstrect)
{
    CAMLparam4(src,srcrect,dst,dstrect);
    SDL_Rect r1,r2;
    r1.x=Int_val(Field(srcrect,0));
    r1.y=Int_val(Field(srcrect,1));
    r1.w=Int_val(Field(srcrect,2));
    r1.h=Int_val(Field(srcrect,3));
    r2.x=Int_val(Field(dstrect,0));
    r2.y=Int_val(Field(dstrect,1));
    r2.w=Int_val(Field(dstrect,2));
    r2.h=Int_val(Field(dstrect,3));
    
    if (SDL_BlitSurface( ML_surf_to_C(src), &r1, ML_surf_to_C(dst), &r2) == -1) 
	fail();
    CAMLreturn0;
}

void MLSDL_fill_rect(value dst, value dstrect, value color)
{
    CAMLparam3(dst,dstrect,color);
    SDL_Rect r;
    r.x=Int_val(Field(dstrect,0));
    r.y=Int_val(Field(dstrect,1));
    r.w=Int_val(Field(dstrect,2));
    r.h=Int_val(Field(dstrect,3));
    if (SDL_FillRect( ML_surf_to_C(dst), &r, Int32_val(color)) == -1)
	fail();
    CAMLreturn0;
}

value MLSDL_display_format(value surf)
{
    CAMLparam1(surf);
    SDL_Surface *csurf;
    
    csurf=SDL_DisplayFormat(ML_surf_to_C(surf));
    if (csurf==NULL)
	fail();
    CAMLreturn(SDL_surf_to_ML_surf(csurf));
}

void MLSDL_display_format_alpha(value surf)
{
    CAMLparam1(surf);
    SDL_Surface *csurf;
    
    csurf=SDL_DisplayFormatAlpha(ML_surf_to_C(surf));
    if (csurf==NULL)
	fail();
    CAMLreturn0;
}


void MLSDL_gl_set_attribute(value attr, value v)
{
    CAMLparam2(attr, v);
    if ( SDL_GL_SetAttribute( Int_val(attr), Int_val(v)) == -1)
	fail();
    CAMLreturn0;
}

void MLSDL_gl_swap_buffers(value u)
{
    CAMLparam1(u);
    SDL_GL_SwapBuffers();
    CAMLreturn0;
}

value MLSDL_gl_get_attribute(value attr)
{
    CAMLparam1(attr);
    int i;
    i=0;
    if(SDL_GL_GetAttribute(Int_val(attr), &i) == -1) fail();
    CAMLreturn(Val_int(i));
}
